import os
import sys

from pypigeonhole_build.dependency import Dependency, INSTALL, DEV, PIP
from pypigeonhole_build.conda_translator import CONDA
import pypigeonhole_build.pip_translator as pip_translator
import pypigeonhole_build.app_version_control as vc
import pypigeonhole_build.dep_manager as dep_manager

curr_dir = os.path.dirname(os.path.realpath(__file__))
__app_name = os.path.basename(curr_dir).replace('_', '-')

# ##############################################################################
# These are the settings for the app.
# ##############################################################################
# 3 digits, major, minor, patch. Keep this line unique.
__app_version = "0.0.3"
vc.bump_version = vc.bump_version_upto10

top_pkg = __app_name.replace('-', '_')
CONDA.env = top_pkg  # _ is easier to copy the word
CONDA.channels = ['defaults', 'psilons', 'psilons-local']  # update channels, if needed.

dependent_libs = [
    # your dependencies here
    # your dependencies here
    Dependency(name='pycryptodome', version='==3.9.8', scope=INSTALL, installer=CONDA),
    Dependency(name='pyyaml', version='==5.3.1', scope=INSTALL, installer=CONDA),
    Dependency(name='pypigeonhole-simple-utils', scope=INSTALL, installer=CONDA),

    Dependency(name='pypigeonhole-build', installer=CONDA),  # default scope DEV, latest version

    Dependency(name='python', version='>=3.6', scope=INSTALL, installer=CONDA),
    Dependency(name='pip', installer=CONDA),  # Without this Conda complains
    Dependency(name='coverage', version='==5.3', installer=CONDA, desc='test coverage'),  # DEV
    Dependency(name='pipdeptree', scope=DEV, installer=PIP),
    Dependency(name='coverage-badge'),  # default to DEV and PIP automatically.
    Dependency(name='twine'),  # uploading to pypi
    Dependency(name='conda-build', installer=CONDA),
    Dependency(name='conda-verify', installer=CONDA),
    Dependency(name='anaconda-client', installer=CONDA),
]

# ##############################################################################
# No need change below, unless you want to customize
# ##############################################################################

# used by setup.py, hide details - how we compute these values.
install_required = pip_translator.get_install_required(dependent_libs)

test_required = pip_translator.get_test_required(dependent_libs)

python_required = pip_translator.get_python_requires(dependent_libs)


def app_name():
    return __app_name


def app_version():
    return __app_version


if __name__ == "__main__":
    dep_manager.main(sys.argv, dependent_libs)



import os

import pypigeonhole_simple_utils.simple.simple_log as slog

logger = slog.get_logger('__name__')


def _chain_with(sep='.', *args):
    if args[0] == '':
        return sep.join((args[1:]))
    else:
        return sep.join(args)


###############################################################################
# Tree structured key value pair store. Value are stored at leaf nodes only.
# Intermediate nodes are for paths only.
###############################################################################
class TreePathSecretDict:
    # we include crypto because we don't want to show secrets in print or log.
    def __init__(self, crypto=None, tree_path_sep='.', secret_prefix='secret='):
        self._tree_path_sep = tree_path_sep
        self._secret_prefix = secret_prefix
        self._crypto = crypto
        self._data_dict = {}

    def get(self, key, default=None):
        nodes = key.split(self._tree_path_sep)

        data = self._data_dict
        for node in nodes:  # walk down tree path nodes
            data = data.get(node, None)  # do not return default here.
            if not data:
                ret = os.environ.get(key)  # if dict does not have it, check env
                if ret:
                    return ret

                return default  # return default only at the end.

        if type(data) is str and data.startswith(self._secret_prefix) and self._crypto:
            value = data[len(self._secret_prefix):]
            return self._crypto.decrypt_secret(value)
        else:
            return data

    def set(self, key, value):
        nodes = key.split(self._tree_path_sep)
        last = len(nodes) - 1
        data = self._data_dict
        for idx, node in enumerate(nodes):
            if idx == last:
                data[node] = value
            else:  # build tree_path down
                next_data = data.get(node, None)
                if not next_data:
                    next_data = self._new_empty_copy()
                    data[node] = next_data

                data = next_data._data_dict

    # dict syntax [], no bomb out with default None
    def __getitem__(self, key):
        return self.get(key, None)

    def __setitem__(self, key, value):
        self.set(key, value)

    def __len__(self):
        tree_path_dict = self.to_dict()
        return len(tree_path_dict)

    def __repr__(self):
        return 'TreePathSecretDict={{sep={}, secret_prefix={}, crypto={}, key_value_pairs={}}}'.format(
            self._tree_path_sep, self._secret_prefix, self._crypto, repr(self._data_dict)
        )

    def merge(self, tree_path_secret_dict: 'TreePathSecretDict'):
        TreePathSecretDict._merge_dicts(self._data_dict, tree_path_secret_dict._data_dict)
        return self

    def add_dict(self, data_dict: dict):
        # assume dict is flattened data, treepath -> value
        self._add_dict('', data_dict)

        return self

    def _add_dict(self, key, data_dict: dict):
        for k, v in data_dict.items():
            kk = _chain_with(self._tree_path_sep, key, k)
            if isinstance(v, dict):
                self._add_dict(kk, v)
            else:
                self.set(kk, v)

    def to_dict(self) -> dict:
        # flatten tree to treepath -> value
        return TreePathSecretDict._to_dict_with_segs('', self._data_dict, self._tree_path_sep)

    def _new_empty_copy(self) -> 'TreePathSecretDict':
        return TreePathSecretDict(self._crypto, self._tree_path_sep, self._secret_prefix)

    # these methods reference internal data dict, so they stay in this class.
    @staticmethod
    def _merge_dicts(dict1: dict, dict2: dict):
        for k, v in dict2.items():
            if k in dict1:
                u = dict1[k]
                if isinstance(u, TreePathSecretDict) and isinstance(v, TreePathSecretDict):
                    TreePathSecretDict._merge_dicts(u._data_dict, v._data_dict)
                else:
                    if isinstance(u, TreePathSecretDict) and not isinstance(v, TreePathSecretDict):
                        raise Exception('not allowed to override path: {} with value {}'.format(u, v))
                    elif not isinstance(u, TreePathSecretDict) and isinstance(v, TreePathSecretDict):
                        raise Exception('not allowed to override value: {} with path: {}'.format(u, v))
                    else:
                        dict1[k] = v  # override dict1's value
            else:
                dict1[k] = v  # shallow copy

    @staticmethod
    def _to_dict_with_segs(seg_prefix: str, data_dict: dict, seg_sep) -> dict:
        # this is mind twisting, recursion. This is deep copying until values.
        # do not deep-copy values, that's not the concern here. Do it elsewhere.
        ret = {}
        for k, v in data_dict.items():
            if isinstance(v, TreePathSecretDict):
                deep_dict = TreePathSecretDict._to_dict_with_segs(_chain_with(seg_sep, seg_prefix, k),
                                                                  v._data_dict, seg_sep)
                for kk, vv in deep_dict.items():
                    ret[_chain_with(seg_sep, seg_prefix, kk)] = vv
            else:
                ret[_chain_with(seg_sep, seg_prefix, k)] = v

        return ret


# This is the optional single entry point, similar to singleton. This is used
# mainly by the framework. Callers in general do not need this. They should use
# top level IoC context to access this.
ioc_setting = None  # syntax sugar, set by below


def set_settings(setting_store: TreePathSecretDict):
    global ioc_setting
    ioc_setting = setting_store



import os
import sys
import typing
import atexit
import signal

# from functools import lru_cache as cached

import yaml

import pypigeonhole_config.ioc_conf as ioc_conf
import pypigeonhole_config.ioc_container as ioc_container
import pypigeonhole_simple_utils.extra.simple_yaml as simple_yaml


# @cached(maxsize=None, typed=False)
def get_sys_args_map(list_of_parameters: list):
    ret = {}
    for arg in list_of_parameters:
        if arg.startswith('-E'):
            param_str = arg[2:]
            idx = param_str.find('=')
            if idx > 0:
                key = param_str[0:idx]
                value = param_str[idx+1:]
                ret[key] = value
    return ret


# @cached(maxsize=None, typed=False)
# def get_ap_args_map(args):
#     ret = {}
#     for arg in vars(args):
#         value = getattr(args, arg)
#         ret[arg] = value
#
#     return ret


# a decorator to collect all factories
class IocRegister:
    factory_list = []

    def __call__(self, *args, **kwargs):
        factory = args[0]
        self.factory_list.append(factory)


ioc_register = IocRegister


T = typing.TypeVar('T')


class IocContext:
    def __init__(self, name, crypto=None, sys_args: dict = None, file_sep='-', file_suffix='.yaml'):
        self.name = name  # can be used in many places, such as configuration file, log file, etc
        if sys_args is None:
            self._cmd_args = get_sys_args_map(sys.argv)
        else:
            self._cmd_args = sys_args
        self._ioc_settings = ioc_conf.TreePathSecretDict(crypto)
        self._ioc_container = ioc_container.IocContainer()

        self._file_sep = file_sep
        self._file_suffix = file_suffix
        atexit.register(self.close)
        signal.signal(signal.SIGINT, self.close)

    def load_conf_file(self, conf_folder: str, env=''):
        file_name = self.name + self._file_sep + env + self._file_suffix if env else self.name + self._file_suffix
        conf_file = os.path.join(conf_folder, file_name)
        conf_data = simple_yaml.load_file(conf_file, self._cmd_args)

        self._ioc_settings.add_dict(conf_data)
        self._ioc_settings.add_dict(self._cmd_args)

        # now override with environment variables. Don't do this, use ${} to assign values
        # for k, v in self._ioc_settings.to_dict().items():  # get full tree path of each setting
        #     res = os.environ.get(k)
        #     if res:
        #         self._ioc_settings[k] = res

        ioc_conf.set_settings(self._ioc_settings)  # point to singleton. If you don't need, it's fine.

    def add_settings(self, extra_settings: dict):  # use to add command line overrides
        self._ioc_settings.add_dict(extra_settings)

    def load_conf_from_yaml_str(self, yaml_str: str):  # mainly for testing purpose
        conf_data = yaml.safe_load(yaml_str)
        self._ioc_settings.add_dict(conf_data)
        ioc_conf.set_settings(self._ioc_settings)

    def init(self):
        factories = IocRegister.factory_list
        for f in factories:
            f(self._ioc_container)

        self._ioc_container.init()

    def close(self):
        self._ioc_container.destroy()

    def get_obj(self, oid, cls: type(T)) -> T:
        obj = self._ioc_container.get(oid, cls)
        return typing.cast(cls, obj)

    def get_conf(self, key):
        return self._ioc_settings.get(key)


app_context = None


def set_app_context(context: IocContext):
    global app_context
    app_context = context



from setuptools import setup, find_packages
import pathlib
import sys
import os.path
from os.path import dirname

# To add source folder to the path, otherwise below import would fail.
src_path = os.path.join(dirname(__file__), 'src')
sys.path.append(src_path)

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

import pypigeonhole_config.app_setup as app_setup

# If this is needed during dev by others, cd this folder and run pip install -e .
# This is reusable in normal cases.
setup(name=app_setup.app_name(),
      version=app_setup.app_version(),
      description='Python YAML based Configuration',
      url='https://github.com/psilons/pypigeonhole-config',

      author='psilons',
      author_email='psilons.quanta@gmail.com',

      long_description=README,
      long_description_content_type="text/markdown",
      license="MIT",

      package_dir={'': 'src'},
      packages=find_packages("src", exclude=["test"]),

      python_requires=app_setup.python_required,

      install_requires=app_setup.install_required,

      tests_require=app_setup.test_required,

      extras_require={},
      )



import typing

import pypigeonhole_config.ioc_conf as ioc_conf


class Params:  # syntax sugar so that code is similar to direct constructor call
    def __init__(self, **kwargs):
        self.param_dict = kwargs  # standard dictionary maintains insert order

    def __repr__(self):
        return repr(self.param_dict)


class Registry:
    def __init__(self, oid: str, cls, constr_params: Params, prop_params: Params,
                 initialize, destroy, singleton=True, lazy=False):
        self.oid = oid
        self.cls = cls
        self.constr_params = constr_params
        self.prop_params = prop_params
        self.initialize = initialize
        self.destroy = destroy
        self.singleton = singleton
        self.lazy = lazy

    def __repr__(self):
        ret = "Registry{{oid={}, cls={}, constr_params={}, prop_params={}, " +\
              "initialize={}, destroy={}, singleton={}, lazy= {}}}"
        return ret.format(self.oid, self.cls, self.constr_params, self.prop_params,
                          self.initialize, self.destroy, self.singleton, self.lazy)


# These 2 classes are to distinguish them from literal values / constants
class ObjRef:
    def __init__(self, key):
        self.key = key

    def __repr__(self):
        return 'ObjRef{{key={}}}'.format(self.key)


class ConfRef:
    def __init__(self, key):
        self.key = key

    def __repr__(self):
        return 'ConfRef{{key={}}}'.format(self.key)


T = typing.TypeVar('T')


class IocContainer:
    def __init__(self):
        self._id_registry = {}
        self._id_object = {}

    def register(self, oid: str, cls, constr_params=Params(), prop_params=Params(),
                 initialize=None, destroy=None, singleton=True, lazy=False):
        self._id_registry[oid] = Registry(oid, cls, constr_params, prop_params,
                                          initialize, destroy, singleton, lazy)

    def get(self, oid, cls: type(T)) -> T:
        obj = self._id_object.get(oid, None)
        if obj is None:
            registry = self._id_registry[oid]
            obj = self._get_or_create(registry, False)
            self._set_obj_props(obj, registry)
            if registry.initialize:
                registry.initialize(obj)

        return typing.cast(cls, obj)

    def init(self):
        self._create_objects()
        self._set_props()
        self._init_objects()

    def destroy(self):
        for oid, registry in self._id_registry.items():
            if registry.singleton and registry.destroy:
                obj = self._id_object[oid]
                registry.destroy(obj)

    def _build_dep_trees(self, id_reg_dict: dict):
        # obj -> dependencies, 1 to many
        # dependency used by many objects, 1 to many
        obj_2_parent = {}
        obj_2_child = {}
        for oid, registry in id_reg_dict.items():
            if oid in self._id_object:
                continue

            if registry.lazy or not registry.singleton:
                continue  # we don't pre-create these objects

            dep_list = []  # collect all object references (not literals or settings)
            for k, v in registry.constr_params.param_dict.items():
                if isinstance(v, ObjRef) and v.key not in self._id_object:
                    dep_list.append(v.key)

            obj_2_child[oid] = dep_list

            for dep_oid in dep_list:
                parent_list = obj_2_parent.get(dep_oid, None)
                if not parent_list:
                    parent_list = []
                    obj_2_parent[dep_oid] = parent_list

                parent_list.append(oid)

        return obj_2_child, obj_2_parent

    def _create_objects(self):
        # worst case is linear dependency, like a -> b -> c -> d
        obj_2_child, obj_2_parent = self._build_dep_trees(self._id_registry)
        for oid, registry in self._id_registry.items():
            if oid in self._id_object:
                continue

            if registry.singleton:
                parents = obj_2_parent.get(oid, None)
                if registry.lazy and parents and len(parents) > 0:  # non top level lazy needs to create
                    self._get_or_create(registry, True)
                elif not registry.lazy:
                    self._get_or_create(registry, True)

    def _get_or_create(self, registry: Registry, prebuild: bool, hist=None):
        if hist is None:  # to track dependency path as we walk, prevent cyclic dependency
            hist = set()

        if registry.oid in self._id_object:
            return self._id_object[registry.oid]

        cps = {}
        for k, v in registry.constr_params.param_dict.items():
            if isinstance(v, ObjRef):
                if v.key in hist:
                    raise Exception('cyclic dependency for object {} in path {}'.format(v.key, hist))
                hist1 = hist.copy()
                hist1.add(v.key)
                reg = self._id_registry[v.key]
                if prebuild and reg.singleton:
                    cps[k] = self._get_or_create(reg, prebuild, hist1)  # recursive call
                else:
                    cps[k] = self.get(reg.oid, reg.cls)
            elif isinstance(v, ConfRef):
                cps[k] = ioc_conf.ioc_setting[v.key]
            else:
                cps[k] = v

        res = registry.cls(**cps)  # call constructor

        if registry.singleton:
            self._id_object[registry.oid] = res

        return res

    def _set_props(self):  # set objects without call set_props and init
        for oid, registry in self._id_registry.items():
            obj = self._id_object.get(oid)
            if obj is None:
                continue  # this happens when lazy=true or singleton=false (prototype)
            for k, v in registry.prop_params.param_dict.items():
                if isinstance(v, ObjRef):
                    dereference = self._id_object.get(v.key, None)
                    if not dereference:  # when prototype or lazy
                        dereference = self.get(v.key, self._id_registry[v.key].cls)
                    setattr(obj, k, dereference)
                elif isinstance(v, ConfRef):
                    dereference = ioc_conf.ioc_setting[v.key]
                    setattr(obj, k, dereference)
                else:
                    setattr(obj, k, v)

    def _set_obj_props(self, obj, registry):
        for k, v in registry.prop_params.param_dict.items():
            if isinstance(v, ObjRef):
                dereference = self.get(v.key, registry.cls)
                setattr(obj, k, dereference)
            elif isinstance(v, ConfRef):
                dereference = ioc_conf.ioc_setting[v.key]
                setattr(obj, k, dereference)
            else:
                setattr(obj, k, v)

    def _init_objects(self):
        for oid, registry in self._id_registry.items():
            if registry.initialize:
                obj = self._id_object.get(oid)
                if obj:
                    registry.initialize(obj)


obj_container = None


def set_ioc_container(container: IocContainer):
    global obj_container
    obj_container = container
